/*=============================================================================================
  mesh conversions
    from files with format :
        .msh usually generated by gmsh (http://www.geuz.org/gmsh/)
        .mel used by melina finite element code (http://perso.univ-rennes1.fr/daniel.martin/melina/)
    to files with format :
        .msh usually generated by gmsh (http://www.geuz.org/gmsh/)
        .mel used by melina finite element code http://perso.univ-rennes1.fr/daniel.martin/melina/
        .vtk used by the 3D visualization tool paraview (http://www.paraview.org/)
        .amdba old mesh format used by some applications

the command syntax is one of the following :
       convmesh [-vl]  file_in file_out
       convmesh [-vl]  file_in format_out
       convmesh -h(elp)   (for mini help)
       convmesh -c(redits)   (for credits)
       options : v for verbose, l=f for french (default value) and l=e for english

       convmesh  mesh.mel mesh.vtk     or convmesh mesh.mel vtk        (no verbose, french)
       convmesh -v mesh.msh mesh.mel   or convmesh -v mesh.msh mel     (verbose, french)
       convmesh -e mesh.msh mesh.mel   or convmesh -e mesh.msh mel     (no verbose, english)
       convmesh -ve mesh.msh mesh.mel  or convmesh -ve mesh.msh mel    (verbose, english)

=============================================================================================*/

#include "mesh.h"

using namespace std;

//global variables
extern map<uint,element_type> element_type_list;  //load in init()
extern string context;   //contains the chain of called functions (error utility)
bool verbose;
ofstream debug;


//main
int main(int argc, char** argv)
{   context="main";
    debug.open("debug.dat");

    //initialisation (def of elements, ...)
    init();

    //arguments analysis
    if(argc==1) {cout<<"help :  conv_mesh -h \n";exit(0);}
    string a1=string(argv[1]);uint l1=a1.size();
    if(a1.substr(0,2)=="-h")
      {cout<<"syntax : conv_mesh [option] filemesh_in.ext1 filemesh_out.ext2\n";
       cout<<"         conv_mesh [option] filemesh_in.ext1 ext2\n";
       cout<<"      ext1= msh | mel ; ext2=msh | mel | vtk |amdba\n";
       cout<<"      options : -h(elp) -v(erbose) -c(redit) -f(rancais) -e(nglish)\n";
       exit(0);
    }
    verbose=false;language lg=francais;string f_in,f_out;
    if(a1.substr(0,2)=="-c") {cout<<"credit : E. LUNEVILLE, U.M.A/POEMS, ENSTA_ParisTech\n";exit(0);}
    if(a1[0]=='-')
       {if(l1>1 && a1[1]=='v') verbose=true;
        if(l1>1 && a1[1]=='e') lg=english;
        if(l1>2 && a1[2]=='v') verbose=true;
        if(l1>2 && a1[2]=='e') lg=english;
        if(argc<4) {cout<<"syntax : conv_mesh [option] filemesh_in.ext1 filemesh_out.ext2\n";
                    cout<<"         conv_mesh [option] filemesh_in.ext1 ext2\n";
                    exit(0);}
        f_in=string(argv[2]);
        f_out=string(argv[3]);
       }
    else
       {if(argc<3) {cout<<"syntax : conv_mesh [option] filemesh_in.ext1 filemesh_out.ext2\n";
                    cout<<"         conv_mesh [option] filemesh_in.ext1 ext2\n";
                    exit(0);}
        f_in=string(argv[1]);
        f_out=string(argv[2]);
       }
    //define language and initialize messages
    init_messages(lg);

    //file analysis
    size_t p=f_in.find_last_of('.',f_in.size());
    if(p==string::npos) message_event("err_filein",msg_data(f_in));
    string ext_in=f_in.substr(p+1,f_in.size());
    if(ext_in!="mel" && ext_in!="msh") message_event("err_filein",msg_data(f_in));
    size_t q=f_out.find_last_of('.',f_out.size());
    string ext_out=f_out;
    if(q==string::npos) f_out=f_in.substr(0,p)+"."+ext_out;
    else ext_out=f_out.substr(q+1,f_in.size());
    if(ext_out!="mel" && ext_out!="msh" && ext_out!="vtk" && ext_out!="amdba")
        message_event("err_fileout");
    if(ext_out==ext_in) message_event("err_inout");

    //new mesh from file
    mesh M(f_in.c_str());

    //M.print(debug,true);

    //export mesh
    if(ext_out=="mel")    M.saveto_mel(f_out.c_str(),false);
    if(ext_out=="msh")    M.saveto_msh(f_out.c_str());
    if(ext_out=="vtk")    M.saveto_vtk(f_out.c_str());
    if(ext_out=="amdba")  M.saveto_amdba(f_out.c_str());

    debug.close();
    return 0;
}
