#include "mesh.h"
#include "utils.h"
#include <algorithm>
#include <cmath>

//=============================================================================
//                       i n i t i a l i s a t i o n
//=============================================================================

map<uint,element_type> element_type_list;
map<string,mesh_type> mesh_type_list;
extern bool verbose;
extern ofstream debug;

void init()
{push_context("init");
 element_type_list[1]=element_type(1,1,1,"lineP1","SE01","SEGMENT",3,2,2,2);
    vector<uint> e(2); e[0]=15;
    e[1]=1;element_type_list[1].edges.push_back(e);
    e[1]=2;element_type_list[1].edges.push_back(e);
 element_type_list[8]=element_type(8,1,2,"lineP2","SE02","UNDEF",3,3,2,2);
    e[1]=1;element_type_list[8].edges.push_back(e);
    e[1]=2;element_type_list[8].edges.push_back(e);
 element_type_list[2]=element_type(2,2,1,"triangleP1","TR01","TRIANGLE",5,3,3,3);
    e.resize(3);e[0]=1;
    e[1]=1;e[2]=2;element_type_list[2].edges.push_back(e);
    e[1]=2;e[2]=3;element_type_list[2].edges.push_back(e);
    e[1]=3;e[2]=1;element_type_list[2].edges.push_back(e);
 element_type_list[9]=element_type(9,2,2,"triangleP2","TR02","UNDEF",5,6,3,3);
    e.resize(4);e[0]=8;
    e[1]=1;e[2]=2;e[3]=4;element_type_list[9].edges.push_back(e);
    e[1]=2;e[2]=3;e[3]=5;element_type_list[9].edges.push_back(e);
    e[1]=3;e[2]=1;e[3]=6;element_type_list[9].edges.push_back(e);
 element_type_list[3]=element_type(3,2,1,"quadrangleQ1","QU01","QUADRANGLE",9,4,4,4);
    e.resize(3);e[0]=1;
    e[1]=1;e[2]=2;element_type_list[3].edges.push_back(e);
    e[1]=2;e[2]=3;element_type_list[3].edges.push_back(e);
    e[1]=3;e[2]=4;element_type_list[3].edges.push_back(e);
    e[1]=4;e[2]=1;element_type_list[3].edges.push_back(e);
 element_type_list[10]=element_type(3,2,2,"quadrangleQ2","QU02","UNDEF",9,9,4,4);
    e.resize(4);e[0]=8;
    e[1]=1;e[2]=2;e[3]=5;element_type_list[10].edges.push_back(e);
    e[1]=2;e[2]=3;e[3]=6;element_type_list[10].edges.push_back(e);
    e[1]=3;e[2]=4;e[3]=7;element_type_list[10].edges.push_back(e);
    e[1]=4;e[2]=1;e[3]=8;element_type_list[10].edges.push_back(e);
 element_type_list[4]=element_type(4,3,1,"tetrahedronP1","TE01","TETRAEDRE",10,4,4,4);
    e[0]=2;
    e[1]=1;e[2]=2;e[3]=3;element_type_list[4].edges.push_back(e);
    e[1]=1;e[2]=2;e[3]=4;element_type_list[4].edges.push_back(e);
    e[1]=2;e[2]=3;e[3]=4;element_type_list[4].edges.push_back(e);
    e[1]=1;e[2]=3;e[3]=4;element_type_list[4].edges.push_back(e);
 element_type_list[11]=element_type(11,3,2,"tetrahedronP2","TE02","UNDEF",10,10,4,4);
    e.resize(7);e[0]=9;
    e[1]=1;e[2]=2;e[3]=3;e[4]=5;e[5]=6;e[7]=8;element_type_list[11].edges.push_back(e);
    e[1]=1;e[2]=3;e[3]=4;e[4]=7;e[5]=8;e[7]=9;element_type_list[11].edges.push_back(e);
    e[1]=2;e[2]=3;e[3]=4;e[4]=6;e[5]=8;e[7]=10;element_type_list[11].edges.push_back(e);
    e[1]=1;e[2]=2;e[3]=4;e[4]=5;e[5]=10;e[7]=9;element_type_list[11].edges.push_back(e);
element_type_list[5]=element_type(5,3,1,"hexahedronQ1","HE01","HEXAEDRE",12,8,8,6);
    e.resize(5);e[0]=3;
    e[1]=1;e[2]=2;e[3]=3;e[4]=4;element_type_list[5].edges.push_back(e);
    e[1]=2;e[2]=6;e[3]=7;e[4]=3;element_type_list[5].edges.push_back(e);
    e[1]=1;e[2]=2;e[3]=6;e[4]=5;element_type_list[5].edges.push_back(e);
    e[1]=4;e[2]=3;e[3]=7;e[4]=8;element_type_list[5].edges.push_back(e);
    e[1]=1;e[2]=5;e[3]=8;e[4]=4;element_type_list[5].edges.push_back(e);
    e[1]=5;e[2]=6;e[3]=7;e[4]=8;element_type_list[5].edges.push_back(e);
 element_type_list[6]=element_type(6,3,1,"prismO1","PR01","UNDEF",13,6,6,5);
    e[0]=2;e[1]=1;e[2]=2;e[3]=3;element_type_list[6].edges.push_back(e);
    e[0]=2;e[1]=4;e[2]=5;e[3]=6;element_type_list[6].edges.push_back(e);
    e[0]=3;e[1]=1;e[2]=2;e[3]=5;e[4]=4;element_type_list[6].edges.push_back(e);
    e[0]=3;e[1]=2;e[2]=5;e[3]=6;e[4]=3;element_type_list[6].edges.push_back(e);
    e[0]=3;e[1]=1;e[2]=4;e[3]=6;e[4]=43;element_type_list[6].edges.push_back(e);
 element_type_list[7]=element_type(5,3,1,"pyramidO1","PY01","UNDEF",14,5,5,4);
    e[0]=3;e[1]=1;e[2]=2;e[3]=3;e[4]=4;element_type_list[6].edges.push_back(e);
    e[0]=2;e[1]=1;e[2]=2;e[3]=5;element_type_list[6].edges.push_back(e);
    e[0]=2;e[1]=2;e[2]=3;e[3]=5;element_type_list[6].edges.push_back(e);
    e[0]=2;e[1]=3;e[2]=5;e[3]=4;element_type_list[6].edges.push_back(e);
    e[0]=2;e[1]=1;e[2]=4;e[3]=5;element_type_list[6].edges.push_back(e);

// element_type_list[12]=element_type(12,3,2,"hexahedronQ2",27,8,6,10);
// element_type_list[13]=element_type(13,3,2,"prismO2",18,6,5,0);
// element_type_list[14]=element_type(14,3,2,"pyramidO2",14,5,4,0);
 element_type_list[15]=element_type(15,0,0,"point","PO00","UNDEF",1,1,1,0);

//format mesh type
 mesh_type_list["msh"]=msh;
 mesh_type_list["vtk"]=vtk;
 mesh_type_list["mel"]=mel;
 mesh_type_list["amdba"]=amdba;
 pop_context();
}

//=============================================================================
//                       c l a s s    n o d e
//=============================================================================
 //translate a node
 node & node::translate(const vector<double>& t)
 {uint n=min(t.size(),size_t(3));
  for(uint i=0;i<n;i++) x[i]+=t[i];
  return *this;
 }

 //scale a node N=O+alpha(N-O)
 node & node::scale(double alpha,const vector<double>& o=vector<double>(3,0.))
 {uint n=min(o.size(),size_t(3));
  for(uint i=0;i<n;i++) x[i]=o[i]+alpha*(x[i]-o[i]);
  return *this;
 }

 //dilate a node N=O+lambda.*(N-O)
 node & node::dilate(const vector<double>& lambda,const vector<double>& o=vector<double>(3,0.))
 {uint n=min(o.size(),size_t(3));
  for(uint i=0;i<n;i++) x[i]=o[i]+lambda[i]*(x[i]-o[i]);
  return *this;
 }

 //rotate in X,Y plane a node N=O+rot_theta(N-O)
 node & node::rotateXY(double theta,const vector<double>& o=vector<double>(2,0.))
 { double ct=cos(theta),st=sin(theta);
   double xx=x[0]-o[0],yy=x[1]-o[1];
   x[0]=o[0]+ct*xx-st*yy;
   x[1]=o[1]+st*xx+ct*yy;
  return *this;
 }


//=============================================================================
//                       c l a s s    l i s t _ o f _ n o d e s
//=============================================================================

//find a node by its number
node & list_of_nodes::find(uint i)
{push_context("node::list_of_nodes");
 it_ln it=nodes.find(i);
 if(it==nodes.end()) message_event("out_of_list");
 pop_context();
 return nodes[i];
}

//access to a node by its number
node & list_of_nodes::operator()(uint i)
{return find(i);}

//does a node given by its index is not in the list
bool list_of_nodes::not_in_list(uint i)
{it_ln it=nodes.find(i);
 return (it==nodes.end());
}

//erase a node of the list
void list_of_nodes::erase(uint i)
{it_ln it=nodes.find(i);
 if(it!=nodes.end()) nodes.erase(it);
}

//erase all the nodes of the list
void list_of_nodes::clear()
{nodes.clear();}

//print the list of nodes
ostream & operator<<(ostream & o,const list_of_nodes & L)
{for(cit_ln it=L.begin();it!=L.end();++it)
    cout<<it->second<<" ";
 cout<<"\n";
 return o;
}

//last index of nodes, may be different to the size of nodes
uint list_of_nodes::last_node() const
{uint ln=0;
 for(cit_ln it=nodes.begin();it!=nodes.end();++it)
    ln=max(ln,it->first);
 return ln;
}

//=============================================================================
//                       c l a s s    e l e m e n t
//=============================================================================

element::element(uint n,uint t,const vector<uint> &ns,uint nd, uint ng, uint np)
{num=n;type=t;nodes=ns;num_domain=nd;num_geom=ng;num_part=np;num_order=0;}

ostream & operator<<(ostream & o,const element & E)
   {//language dependances
    string NOD,DOM,BOR,DEL;
     switch (msg::lang)
    {case english : NOD=" nodes : ";DOM="domaine ";BOR=" edge ";DEL=" of elt ";
                    break;
     default : NOD=" noeuds : ";DOM="domain ";BOR=" bord ";DEL=" de l'elt ";
                    break;
    }
    o<<"elt "<<E.num<<" type "<<element_type_list[E.type].name<<NOD;
    for(uint i=0;i<E.nodes.size();i++) o<<E.nodes[i]<<" ";
    o<<DOM<<E.num_domain;
    list<pair<uint,uint> >::const_iterator it;
    for(it=E.boundary_element.begin();it!=E.boundary_element.end();++it)
         o<<BOR<<it->second<<DEL<<it->first;
    o<<"\n";
    return o;
   }

//create a list of the edge elements, incomplete data (no index)
vector<element> element::edges() const
{element_type & et=element_type_list[type];
 vector<element> edge_elts(et.number_of_edges);
 for(uint k=0;k<et.number_of_edges;k++)                  //loop on edges of the element
    {uint typk=et.edges[k][0];                           //type of the k th edge
     element_type & etk=element_type_list[typk];         //element type of the k h edge
     edge_elts[k].type=typk;                             //fill properties of the kth edge element
     edge_elts[k].num=k+1;                               //local index !!!!
     edge_elts[k].nodes.resize(etk.number_of_nodes);
     for(uint i=1;i<=etk.number_of_nodes;i++)            //list of nodes of the kth edge element
         edge_elts[k].nodes[i-1]=nodes[et.edges[k][i]-1];
     edge_elts[k].boundary_element.push_back(pair<uint,uint>(num,k+1));
    }
 return edge_elts;
}

//return the string index of an element : list of ordered vertices index ("xx y zzz")
//useful for comparison (geometrical) of two elements
string element::string_index()
{string s;
 uint nbv=element_type_list[type].number_of_vertices;
 vector<uint> N(nbv);
 for(uint i=0;i<nbv;i++) N[i]=nodes[i];
 sort(N.begin(),N.end());
 for(uint i=0;i<nbv;i++) s+=int2str(N[i])+" ";
 return s;
}

//test if an element has a boundary with an other element
bool element::is_adjacent_to(const element& elt)
{return common_edge(*this,elt).first!=0;
}

//find if two elements have a common edge
pair<uint,uint> common_edge(const element& elt1,const element& elt2)
{vector<element> edg1=elt1.edges();
 vector<element> edg2=elt2.edges();
 for(uint i=0;i<edg1.size();i++)
    for(uint j=0;j<edg2.size();j++)
       if(edg1[i].string_index()==edg2[j].string_index()) return pair<uint,uint>(i,j);
 return pair<uint,uint>(0,0);
}


//=============================================================================
//                       c l a s s    l i s t _ o f _ e l e m e n t s
//=============================================================================

//find an element by its number
element & list_of_elements::find(uint i)
{push_context("list_of_elements::find");
 it_le it=elements.find(i);
 if(it==elements.end()) message_event("out_of_list2",msg_data(int(i)));
 pop_context();
 return elements[i];
}

const element & list_of_elements::find(uint i) const
{push_context("list_of_elements::find const");
 cit_le it=elements.find(i);
 if(it==elements.end()) message_event("out_of_list2",msg_data(int(i)));
 pop_context();
 return it->second;
}

//access to a element by its number
element & list_of_elements::operator()(uint i)
{return find(i);}
const element & list_of_elements::operator()(uint i) const
{return find(i);}

//erase a element of the list
void list_of_elements::erase(uint i)
{it_le it=elements.find(i);
 if(it!=elements.end()) elements.erase(it);
}

//erase all the elements of the list
void list_of_elements::clear()
{elements.clear();}

//print the list of elements
ostream & operator<<(ostream & o,const list_of_elements & L)
{for(cit_le it=L.begin();it!=L.end();++it) o<<it->second<<" ";
 o<<"\n";
 return o;
}

//last index of elements, may be different to the size of elements
uint list_of_elements::last_elt() const
{uint le=0;
 for(cit_le it=elements.begin();it!=elements.end();++it)
    le=max(le,it->first);
 return le;
}

//=============================================================================
//           c l a s s    p h y s i c a l _ d o m a i n
//=============================================================================

ostream & operator << (ostream &o, const physical_domain &d)
{o<<"domaine "<<d.num<<" : "<<d.name;
 if(d.parents.size()>0)
   {o<<" de parent ";
    for(uint i=0;i<d.parents.size();i++) o<<" "<<d.parents[i]->name;
   }
 if(d.multidim) o<<" multidimensionnel, de dimension maximum "<<d.dim;
 else o<<" de dimension "<<d.dim;
 uint n=d.elements.size();
 o<<" nombre d'lments "<<n;
 if(n>0)
   {o<<"\nListe des lments : (pointeur mesh="<<d.pmesh<<")\n";
    const mesh &M=*d.pmesh;
    for(uint i=0;i<n;i++)  o<<M.elements_list(d.elements[i]);
   }
 return o;
}

ostream & operator<<(ostream & o,const dom_info & di)
{string DOM,ELT,BOU,OF,NN,IG,ND,CRA,CRD,SID;
 switch (msg::lang)
    {case english : DOM=" domain ";ELT=" elements)";BOU=" boundary of ";OF=", of ";NN=" new name : ";
                    IG=" ignored domain ";ND=" new definition"; CRA=" crack";CRD=" cracked";SID=" side of"; break;
     default : DOM=" domaine ";ELT=" lements)";BOU=" frontire de ";OF=", de ";NN=" nouveau nom : ";
               IG=" domaine ignor ";ND=" nouvelle definition";CRA=" fissure";CRD=" fissur";SID=" cot de";break;
    }
 string sdim[4];sdim[0]="0D (point)";sdim[1]="1D";sdim[2]="2D";sdim[3]="3D";
 if(!di.new_def)
   {o<<"\n"<<di.domain->name<<DOM<<sdim[di.domain->dim]<<" ("<<di.domain->elements.size()<<ELT;
    if(di.domain->dim < di.domain->pmesh->dim)
      {if(di.domain->parents.size()>0) o<<BOU<<di.domain->parents[0]->name;
       for(uint i=1;i<di.domain->parents.size();i++) o<<OF<<di.domain->parents[i]->name;
      }
    if(di.keep && di.new_name!="" && di.new_name!=di.domain->name) o<<NN<<di.new_name;
    if(di.cracked) o<<CRD;
    if(di.crack) o<<CRA;
    if(!di.keep) o<<IG;
   }
   else
     {if(di.cracked) o<<"\n"<<di.new_name<<SID<<di.domain->name<<BOU<<di.parent_domain->name<<ND<<CRA;
      else         o<<"\n"<<di.new_name<<DOM<<di.domain->name<<BOU<<di.parent_domain->name<<ND;
     }
 return o;
}

//=============================================================================
//                       c l a s s    m e s h
//=============================================================================

//copy constructor
mesh::mesh(const mesh &M)
{dim=M.dim; loadfrom=M.loadfrom;
 name="copy_of_"+M.name;
 title=M.title;
 xmin=M.xmin;xmax=M.xmax;ymin=M.ymin;ymax=M.ymax;zmin=M.zmin;zmax=M.zmax;
 xv=M.xv;yv=M.yv;zv=M.zv;
 nodes_list=M.nodes_list;
 elements_list=M.elements_list;
 //copy of domains (be care with pointer)
 domains_list=M.domains_list;
 for(it_ld it=domains_list.begin();it!=domains_list.end();++it)
    {it->second.pmesh=this;
     //update parents
     for(uint i=0;i<it->second.parents.size();i++)
        if(it->second.parents[i]!=NULL)
           it->second.parents[i]=&domains_list[it->second.parents[i]->num];
    }
}

//constructor from a regular mesh of a segment in x component (nby=0)
mesh::mesh(double x1,double x2, uint nbx,
           string& dom_name,vector<string>& boundary_name)
{push_context("mesh::mesh (regular mesh of a segment)");
 if(nbx==0) message_event("err_constructor_1D");
 if(dom_name=="") dom_name="SEGMENT["+float2str(x1)+","+float2str(x2)+"]";
 name=dom_name;
 title="segment mesh tool : "+dom_name;
 segment_mesh(nbx,dom_name,boundary_name);
 scale(x2-x1);
 vector<double> d(1);d[0]=x1;
 translate(d);
 pop_context();
}

//constructor from a regular mesh of a 2D box
mesh::mesh(double x1,double x2, uint nbx,
           double y1,double y2, uint nby,
           string& dom_name,vector<string>& boundary_name)
{push_context("mesh::mesh (regular mesh of a 2D box)");
 if(nbx*nby==0)
    {msg_data md;md<<int(nbx);md<<int(nby);
     message_event("err_constructor_2D",md);
    }
 if(dom_name=="") dom_name="RECTANGLE["+float2str(x1)+","+float2str(x2)+
                                  "]x["+float2str(y1)+","+float2str(y2)+"]";
 name=dom_name;
 title="rectangle mesh tool : "+dom_name;
 square_mesh(nbx,nby,dom_name,boundary_name);
 vector<double> d(2);d[0]=x2-x1;d[1]=y2-y1;
 dilate(d);
 d[0]=x1;d[1]=y1;
 translate(d);
 pop_context();
}

//regular mesh of the unit segment
void mesh::segment_mesh(uint nbx,string& dom_name,
                    vector<string>& boundary_name)
{push_context("mesh::segment_mesh");
 dim=1;
 loadfrom=generated;
 xmin=0;xmax=1.;xv=true;yv=false;zv=false;
 double dx=1./nbx,x=0.;
 //nodes generation
 for(uint i=0;i<=nbx;i++,x+=dx) nodes_list.add(i+1,node(x));
 //elements generation
 vector<uint> num(2);
 for(uint l=1;l<=nbx;l++)
    {num[0]=l;num[1]=l+1;
     elements_list.add(l,element(l,1,num,0));
    }
 //physical_domains generation
 domains_list[0]=physical_domain(0,dom_name,this,1);
 for(uint l=1;l<=nbx;l++)domains_list[0].elements.push_back(l);
 uint nbel=nbx;
 //boundary domains (end points of the segment)
 if(boundary_name[0]!="")
   {nbel++;
    element b1(nbel,15,vector<uint>(1,1),1,0,0);
    b1.boundary_element.push_back(pair<uint,uint>(1,1));
    elements_list.add(nbel,b1);
    domains_list[1]=physical_domain(1,boundary_name[0],this,0);
    domains_list[1].parents.push_back(&domains_list[0]);
    domains_list[1].elements.push_back(nbel);
   }
 if(boundary_name[1]!="")
   {nbel++;
    element b2(nbel,15,vector<uint>(1,nbx+1),1,0,0);
    b2.boundary_element.push_back(pair<uint,uint>(nbx,2));
    elements_list.add(nbel,b2);
    domains_list[2]=physical_domain(1,boundary_name[1],this,0);
    domains_list[2].parents.push_back(&domains_list[0]);
    domains_list[2].elements.push_back(nbel);
   }
 pop_context();
}

//regular mesh of the unit square
void mesh::square_mesh(uint nbx,uint nby,string& dom_name,
                       vector<string>& boundary_name)
{push_context("mesh::square_mesh");
 dim=2;
 loadfrom=generated;
 xmin=0;xmax=1.;ymin=0.;ymax=1.;xv=true;yv=true;zv=false;
 double dx=1./nbx,x=0.;
 double dy=1./nby,y=0.;
 //nodes generation
 uint k=0;
 for(uint j=0;j<=nby;j++,y+=dy)
    {
     for(uint i=0;i<=nbx;i++,x+=dx)
        {k++;
         nodes_list.add(k,node(x,y));
        }
     x=0.;
    }
 //elements generation
 vector<uint> num(3);uint nbel=0;
 for(uint j=1;j<=nby;j++)
    for(uint i=1;i<=nbx;i++)
       {uint p=i+(j-1)*(nbx+1);
        num[0]=p;num[1]=p+1;num[2]=p+nbx+1;
        nbel++;
        elements_list.add(nbel,element(nbel,2,num,0));
        num[0]=p+nbx+2;num[1]=p+nbx+1;num[2]=p+1;
        nbel++;
        elements_list.add(nbel,element(nbel,2,num,0));
    }
 //physical_domains generation
 domains_list[0]=physical_domain(0,dom_name,this,2);
 for(uint l=1;l<=nbel;l++)domains_list[0].elements.push_back(l);
 //boundary domains
 for(uint n=1;n<=4;n++)
 if(boundary_name[n-1]!="")  //boundary y=0, x=1, y=1 and x=0
   {domains_list[n]=physical_domain(n,boundary_name[n-1],this,1);
    domains_list[n].parents.push_back(&domains_list[0]);
    uint beg=0,step=0,nbe=0,art=0,bege=0,stepe=0;
    switch (n)
      {case 1 :nbe=nbx;beg=1;step=1;art=1;bege=1;stepe=2;break;
       case 2 :nbe=nby;beg=nbx+1;step=nbx+1;art=3;bege=2*nbx;stepe=2*nbx;break;
       case 3 :nbe=nbx;beg=nby*(nbx+1)+1;step=1;art=2;bege=2*(nby-1)*nbx+2;stepe=2;break;
       case 4 :nbe=nby;beg=1;step=nbx+1;art=3;bege=1;stepe=2*nbx;break;
      }
    for(uint l=1;l<=nbe;l++)
       {nbel++;
        vector<uint> nbo(2);nbo[0]=beg;nbo[1]=beg+step;beg+=step;
        element b(nbel,1,nbo,1);
        b.boundary_element.push_back(pair<uint,uint>(bege,art));
        elements_list.add(nbel,b);
        domains_list[n].elements.push_back(nbel);
        bege+=stepe;
       }

   }
 pop_context();
}

//mesh of a unit crown (internal circle of radius 1)
// nbs : number of segment on the internal circle
// nbl : number of layers  (external radius = 1+nbl*sl)
// sl  : size of the layer
// dom_name : name of the sutface domain
// boundary_name : name of the internal and the external boundaries
//void mesh::crown_mesh(uint nbs,uint nbl,double sl, string& dom_name=string(""),
//                   vector<string>& boundary_name=vector<string>(2,string("")))
//{
//}

//detect null coordinates of the mesh
void mesh::null_coordinates()
{xv=!(xmin==0 && xmax==0);yv=!(ymin==0 && ymax==0);zv=!(zmin==0 && zmax==0);}

//mesh loading
void mesh::loadfromfile(const char * filename)
{push_context("mesh::loadfromfile");
//filename analysis
 name=filename;
 loadfrom=msh;
 xmin=xmax=ymin=ymax=zmin=zmax=0.;
 size_t p=name.find_last_of('.',name.size());
 if(p==string::npos) message_event("err_fileext",msg_data(name));
 string ext=name.substr(p+1);
 string rn=name.substr(0,p);
 //test if file exists
fstream ff;
ff.open(filename,fstream::in);
if(ff.fail()) message_event("open_fail",msg_data(name));
ff.close();
//call the specific loader
if(mesh_type_list.find(ext)==mesh_type_list.end()) message_event("err_fileext2",msg_data(ext));
if(verbose) message_event("info_read",msg_data(filename));
switch(mesh_type_list[ext])
  {case msh : loadfrom_msh(filename);break;
   case mel : loadfrom_mel(filename);break;
   case vtk : loadfrom_vtk(filename);break;
   case amdba : loadfrom_amdba(filename);break;
   case generated : break;
  }
 //print general info
 if(verbose) print(cout,false);
 pop_context();
}

//translate a mesh
mesh & mesh::translate(const vector<double>& t)
{push_context("mesh::translate");
 for(it_ln it=nodes_list.begin();it!=nodes_list.end();++it)
     it->second.translate(t);
 //update the bounding box
 uint n=t.size();
 if(n>0) {xmin=xmin+t[0];xmax=xmax+t[0];}
 if(n>1) {ymin=ymin+t[1];ymax=ymax+t[1];}
 if(n>2) {zmin=zmin+t[2];zmax=zmax+t[2];}
 pop_context();
 return *this;
}

//scale a mesh (isotrope scaling)
mesh & mesh::scale(double alpha, const vector<double>& o)
{push_context("mesh::scale");
 for(it_ln it=nodes_list.begin();it!=nodes_list.end();++it)
     it->second.scale(alpha,o);
 //update the bounding box
 uint n=o.size();
 if(n>0) {xmin=o[0]+alpha*(xmin-o[0]);xmax=o[0]+alpha*(xmax-o[0]);}
 if(n>1) {ymin=o[1]+alpha*(ymin-o[1]);ymax=o[1]+alpha*(ymax-o[1]);}
 if(n>2) {zmin=o[2]+alpha*(zmin-o[2]);zmax=o[2]+alpha*(zmax-o[2]);}
 pop_context();
 return *this;
}

//dilate a mesh (different scaling along the axis)
mesh & mesh::dilate(const vector<double>& lambda, const vector<double>& o)
{push_context("mesh::dilate");
 for(it_ln it=nodes_list.begin();it!=nodes_list.end();++it)
     it->second.dilate(lambda,o);
 //update the bounding box
 uint n=o.size();
 if(n>0) {xmin=o[0]+lambda[0]*(xmin-o[0]);xmax=o[0]+lambda[0]*(xmax-o[0]);}
 if(n>1) {ymin=o[1]+lambda[1]*(ymin-o[1]);ymax=o[1]+lambda[1]*(ymax-o[1]);}
 if(n>2) {zmin=o[2]+lambda[2]*(zmin-o[2]);zmax=o[2]+lambda[2]*(zmax-o[2]);}
 pop_context();
 return *this;
}

//rotate in X,Y plane a mesh
mesh & mesh::rotateXY(double theta, const vector<double>& o=vector<double>(2,0))
{push_context("mesh::rotateXY");
 for(it_ln it=nodes_list.begin();it!=nodes_list.end();++it)
     it->second.rotateXY(theta,o);
 //update the 2D bounding box
 node S1(xmin,ymin);S1.rotateXY(theta,o);
 xmin=S1(1);xmax=S1(1);ymin=S1(2);ymax=S1(2);
 node S2(xmin,ymax);S2.rotateXY(theta,o);
 xmin=min(xmin,S2(1));ymin=min(ymin,S2(2));xmax=max(xmax,S2(1));ymax=max(ymax,S2(2));
 node S3(xmax,ymin);S3.rotateXY(theta,o);xmin=S1(1);ymin=S1(2);
 xmin=min(xmin,S3(1));ymin=min(ymin,S3(2));xmax=max(xmax,S3(1));ymax=max(ymax,S3(2));
 node S4(xmax,ymax);S4.rotateXY(theta,o);xmin=S1(1);ymin=S1(2);
 xmin=min(xmin,S4(1));ymin=min(ymin,S4(2));xmax=max(xmax,S4(1));ymax=max(ymax,S4(2));
 pop_context();
 return *this;
}

//find a domain by its name
physical_domain* mesh::find_domain(const string & na)
{push_context("mesh::find_domain");
 physical_domain *pd=NULL;
 typedef map<uint,physical_domain>::iterator it_pd;
 for(it_pd it=domains_list.begin();it!=domains_list.end();++it)
    if(it->second.name==na) {pd= &(it->second);break;}
 pop_context();
 return pd;
 }

//merge 2 meshes; physical domains are merged if they have the same name
//there is no elimination of nodes having the same coordinates!!!
mesh & mesh::operator+=(const mesh &M)
{push_context("mesh::operator +=");
 typedef list<pair<uint,uint> >::iterator it_pp;
 typedef map<uint,physical_domain>::const_iterator cit_pd;
 typedef vector<uint>::iterator it_vu;
 uint lastnode=nodes_list.size();uint k=lastnode;
 uint lastelt=elements_list.size();uint l=lastelt;
 //dernier numero de domaine = le plus grand
 uint lastdom=last_dom();
 //update name ...
 name+=" + "+M.name;
 title=name;
 loadfrom=generated;
 //update the bounding box
 xmin=min(xmin,M.xmin);xmax=max(xmax,M.xmax);
 ymin=min(ymin,M.ymin);ymax=max(ymax,M.ymax);
 zmin=min(zmin,M.zmin);xmax=max(zmax,M.zmax);
 xv=xv||M.xv;yv=yv||M.yv;zv=zv||M.zv;
 //add nodes
 for(cit_ln it=M.nodes_list.begin();it!=M.nodes_list.end();++it)
    {k++;nodes_list.add(k,it->second);}
 //add elements
 for(cit_le it=M.elements_list.begin();it!=M.elements_list.end();++it)
    {element e=it->second;
     //update element nodes list (tranlate the index)
     for(uint j=0;j<e.nodes.size();j++) e.nodes[j]+=lastnode;
     //update boundary informations
     for(it_pp jt=e.boundary_element.begin(); jt!=e.boundary_element.end();++jt)
         jt->first+=lastelt;
     l++;
     e.num=l;
     elements_list.add(l,e);}
 //add physical domains with merging
 for(cit_pd it=M.domains_list.begin();it!=M.domains_list.end();++it)
    {const physical_domain& pd=it->second;
     physical_domain* pp=find_domain(pd.name);
     if(pp!=NULL) //merge domain
       for(uint i=0;i<pd.elements.size();i++)
             pp->elements.push_back(pd.elements[i]+lastelt);
     else         //add domain
       {uint nd=pd.num+lastdom+1;
        domains_list.insert(pair<uint,physical_domain>(nd,pd));
        physical_domain & pn=domains_list[nd];
        //update data
        pn.pmesh=this;
        for(uint i=0;i<pn.elements.size();i++) pn.elements[i]+=lastelt;
        for(uint i=0;i<pn.parents.size();i++)    //fails if parent is read after child !!!
           {pp=find_domain(pn.parents[i]->name);
            if(pp!=NULL && pp!=&pn) pn.parents[i]=pp;
           }
       }
    }
 pop_context();
 return *this;
}

//print mesh data
void mesh::print(ostream & o,bool detail) const
{push_context("mesh::print");
 //language dependances
  string MA,DIM,NBN,NBE,LIN,LIE,LID,LIED,EDG;
  switch (msg::lang)
    {case english : MA="\nMesh ";DIM="dimension ";NBN=" number of nodes ";NBE=" number of elements ";
                    LIN="list of the nodes :\n";LIE="list of the elements :";LID="list of the domains :\n";
                    LIED="list of edges :\n";EDG="edge ";
                    break;
     default : MA="\nMAillage ";DIM="dimension ";NBN=" nombre de noeuds ";NBE=" nombre d'lments ";
                    LIN="liste des noeuds :\n";LIE="liste des elements :";LID="liste des domaines :\n";
                    LIED="liste des faces ou artes :\n";EDG="face ou arte ";
                    break;
    }
 //print
 o<<MA<<name<<" : ";
 o<<DIM<<dim<<NBN<<nodes_list.size()<<NBE<<elements_list.size()<<"\n";
 if(detail)
   {o<<LIN;   //print list of nodes
     for(cit_ln it=nodes_list.begin();it!=nodes_list.end();++it)
        {node N=it->second;
         o<<it->first<<" : "<<N<<"\n";
        }
      o<<LIE<<"\n"<<elements_list; //print list of elements
      o<<LID; //print list of domains
      map<uint,physical_domain>::const_iterator it;
      for(it=domains_list.begin();it!=domains_list.end();++it)
         {o<<it->second<<"\n";}
      //print list of edges
      if(edges_list.size()!=0)
        {o<<LIED;
         map<string,element>::const_iterator jt;
         for(jt=edges_list.begin();jt!=edges_list.end();++jt)
            o<<EDG<<jt->first<<" : "<<jt->second;
        }
   }
 pop_context();
}

//validation before opening a mesh savefile
bool mesh::valid_savefile(string & filename,const string& ext,bool silent)
{push_context("mesh::valid_savefile");
 string savename=name;
 if(filename!="") savename=filename;
 savename=replace_ext(savename,ext);
 if(FileExists(savename) && !silent)
   {string rep="n";
    message_event("file_confirmation",msg_data(savename));
    cin>>rep;
    if(rep=="n") {pop_context();return false;}
   }
 if(verbose) message_event("info_save",msg_data(savename));
 filename=savename;
 pop_context();
 return true;
}

//return the high index of physical domain of a mesh
uint mesh::last_dom() const
{uint lastd=0;
 typedef map<uint,physical_domain>::const_iterator cit_pd;
 for(cit_pd it=domains_list.begin();it!=domains_list.end();++it)
     lastd=max(lastd,it->first);
 return lastd;
}

//make the list of all edges of the mesh
void mesh::make_edges_list()
{//loop over the volumic elements
 map<string,element>::iterator itse;
 uint edge_index=0;
 for(it_le it=elements_list.begin();it!=elements_list.end();++it)
    {element& e=it->second;
     if(element_type_list[e.type].dim==dim)  //volumic element
       {vector<element> edge_elts=e.edges();
        for(uint k=0;k<edge_elts.size();k++)
           {element &ek=edge_elts[k];
            string sk=ek.string_index();//find if the element belongs to the edges_list
            itse=edges_list.find(sk);
            if(itse==edges_list.end())  //add this element to the edges list
               {edge_index++;ek.num=edge_index;
                edges_list.insert(pair<string,element>(sk,ek));
               }
            else         //update boundary_element info of the the found element
               itse->second.boundary_element.push_back(pair<uint,uint>(e.num,k+1));
           }
       }
    }

}


//tranform a subdomain (dim=mesh.dim-1) into a crack
//two cases are possible :
//  - the case of a subdomain immersed in one volumic domain
//  - the case of a subdomain which separate two volumic domain
//does not work if the subdomain is a boundary of more than 2 volumic domains
// ATTENTION : the edges_list of the mesh is not modified !!!!

void mesh::crack_sub_domain(physical_domain &pd)
{uint lastdom=last_dom();
 uint lastnode=last_node();
 uint lastelt=last_elt();

 //create the list of all edges if not exist
 if(edges_list.size()==0) make_edges_list();
 //print(debug,true);

 //create cracked subdomains
 map<uint,uint> ptoc;    //store the link between parent and child
 for(uint i=0;i<pd.parents.size();i++)
    {string napd=ignore_quotation_mark(pd.name)+"_"+ignore_quotation_mark(pd.parents[i]->name);
     lastdom++;
     domains_list[lastdom]=physical_domain(pd);     //new physical domain from copy of the original one
     domains_list[lastdom].num=lastdom;             //update data
     domains_list[lastdom].name=napd;
     domains_list[lastdom].elements.clear();
     domains_list[lastdom].parents=vector<physical_domain *>(1,pd.parents[i]);
     ptoc[pd.parents[i]->num]=lastdom;
    }

 //make the extended domain of the subdomain
 //that is the set of all elements which have at least one node belonging to the subdomain
 //  - this set is split in two parts according to the side of the crack
 //    explicety in the case of two parent domains or implicitely in the case of one parent domain (hard case)
 //  - for each element it is indicated the side where it is located
 //we use a map of side indexed by the element index

 uint nbel=pd.elements.size();      //number of elements of the subdomain
 map<uint,uint> sd_nodes;           //list of nodes of the subdomain (maps old node index to new node index)
 map<uint,uint> ext_domain;         //list of elements of the extended domain (maps elt index to its side)
 map<uint,uint>::iterator ite,jte,itn,endit; //iterator declaration
 bool firstside=true;
 uint s1,s2;

 //deal with elements which have an edge on the subdomain
 //we create the list of the nodes belonging to the subdomain
 for(uint i=0;i<nbel;i++)
    {element& elt0=elements_list(pd.elements[i]);
    //create new nodes
     for(uint j=0;j<elt0.nodes.size();j++)                 //loop on element nodes
        {map<uint,uint>::iterator itn=sd_nodes.find(elt0.nodes[j]);
         if(itn==sd_nodes.end())
           {node ni=nodes_list(elt0.nodes[j]);            //create a copy of node
            lastnode++;nodes_list.add(lastnode,ni);       //add to the global list of nodes
            sd_nodes[elt0.nodes[j]]=lastnode;             //add to the list of nodes
           }
        }
     list<pair<uint,uint> >::const_iterator it=elt0.boundary_element.begin();
     if(elt0.boundary_element.size()!=2) message_event("err_crack1");  //only one parent -> error
     uint par1=it->first;element& ep1=elements_list(par1);
     it++;uint par2=it->first;element& ep2=elements_list(par2);
     if(ep1.num_domain!=ep2.num_domain) //two domain parents
       {uint md=min(ep1.num_domain,ep2.num_domain);
        if(ep1.num_domain==md) {s1=1;s2=2;} else {s1=2;s2=1;}
       }
     else   //unique domain parent
       {if(firstside) {s1=1;s2=2;firstside=false;}       //firsttime, choose arbitrary the sides
        else
           {//find if the element ep1 is adjacent to an element of side 1 or 2
            ite=ext_domain.begin();
            endit=ext_domain.end();
            while(ite!=endit)
                 {element& eite=elements_list(ite->first);
                  if(ite->second!=0 && ep1.is_adjacent_to(eite))
                    {s1=ite->second;ite=endit;}
                  else ite++;
                 }
            ite=ext_domain.begin();
            while(ite!=endit)
                 {element& eite=elements_list(ite->first);
                  if(ite->second!=0 && ep2.is_adjacent_to(eite))
                    {s2=ite->second;ite=endit;}
                  else ite++;
                 }
            }
        }
     ext_domain.insert(pair<uint,uint>(par1,s1));
     ext_domain.insert(pair<uint,uint>(par2,s2));
    }

 //find the other elements which encounter the subdomain to be cracked
 //these elements are adjacent to one of the previous list
 //and have at least a node on the subdomain
 bool newlist=true;
 while(newlist)
 {map<uint,uint> new_list;
  for(ite=ext_domain.begin();ite!=ext_domain.end();++ite)
     {uint indelt=ite->first;                            //global index of the element
      element & elt=elements_list(indelt);               //the element itself
      vector<element> elt_edges=elt.edges();             //edges of the element (local)
      for(uint j=0;j<elt_edges.size();j++)               //loop on the edges of elt (local)
         {string sj=elt_edges[j].string_index();         //edge string index
          element& eltj=edges_list[sj];                  //edge of element (global)
          bool oncrack=false;uint n=0;
          while(n<eltj.nodes.size() && !oncrack)         //test if the edge has at least a node on the crack
               if(sd_nodes.find(eltj.nodes[n])!=sd_nodes.end()) oncrack=true;else n++;
          if(oncrack)
            {list<pair<uint,uint> >::const_iterator kte;    //loop on parent of the edge j
             for(kte=eltj.boundary_element.begin();kte!=eltj.boundary_element.end();++kte)
                if(kte->first!=indelt &&                                            //not the element elt itself
                   ext_domain.find(kte->first)==ext_domain.end() &&                 //not in the list ext_domain
                   new_list.find(kte->first)==new_list.end() )                      //not in the new_list
                   new_list.insert(pair<uint,uint>(kte->first,ext_domain[indelt])); //insert elt in new_list
            }
         }
     }  //end for(ite=
   //end while
   newlist=new_list.size()>0;
   if(newlist) ext_domain.insert(new_list.begin(),new_list.end());    //add new_list to the ext_domain list
 }

//complete the side information for element which have no side defined
 endit=ext_domain.end();
 for(ite=ext_domain.begin();ite!=endit;++ite)
    {if(ite->second==0)
        {element & elt=elements_list(ite->first);
         jte=ext_domain.begin();
         while(jte!=endit)
             {if(jte!=ite && jte->second!=0)
                {if(elt.is_adjacent_to(elements_list(jte->first)))
                    {ext_domain[ite->first]=jte->second;jte=endit;}
                }
              else jte++;
             }
        }
    }

//    debug<<"\nlist of elements intersecting the crack\n";
//    for(ite=ext_domain.begin();ite!=endit;++ite)
//       {debug<<"elt "<<ite->first<<" side "<<ite->second<<"\n";
//        element& el=elements_list(ite->first);
//        debug<<"    elt nodes : ";
//        for(uint i=0;i<el.nodes.size();i++)
//           debug<<el.nodes[i]<<"->"<<nodes_list(el.nodes[i])<<" ";
//        debug<<"\n";
//       }

//now crack the domain
//create from the original subdomain new nodes and two new subdomains
//and update boundary informations
 for(uint i=0;i<nbel;i++)          //loop on the elements of the original subdomain
    {element& elt0=elements_list(pd.elements[i]);
     //create two new boundary elements
     vector<element> newelts(2);
     for(uint j=0;j<2;j++) newelts[j]=elt0;               //copy elt0
     for(uint i=0;i<newelts[1].nodes.size();i++)          //update the node of the second elt
         newelts[1].nodes[i]=sd_nodes[elt0.nodes[i]];
     //update boundary informations, global index, ...
     list<pair<uint,uint> >::iterator itb=elt0.boundary_element.begin();
     vector<pair<uint,uint> > be(2);
     be[0]=*itb;itb++;be[1]=*itb;
     vector<uint> l(2);
     if(ext_domain[be[0].first]==1) {l[0]=0;l[1]=1;} else {l[0]=1;l[1]=0;}
     for(uint j=0;j<2;j++)
        {uint nudom=ptoc[elements_list(be[l[j]].first).num_domain];
         newelts[j].num_domain=nudom;
         newelts[j].boundary_element.clear();
         newelts[j].boundary_element.push_back(be[l[j]]);
         lastelt++;newelts[j].num=lastelt;
         elements_list.add(lastelt,newelts[j]);
         domains_list[nudom].elements.push_back(lastelt);
        }
    }

//update node indexes of element of ext_domain on side 2
 for(ite=ext_domain.begin();ite!=endit;++ite)
    {if(ite->second==2)
       {element & elt=elements_list(ite->first);
        for(uint j=0;j<elt.nodes.size();j++)
           {itn=sd_nodes.find(elt.nodes[j]);
            if(itn!=sd_nodes.end()) elt.nodes[j]=itn->second;
           }
       }
    }

 print(debug,true);
}


