#ifndef MESSAGE_H_INCLUDED
#define MESSAGE_H_INCLUDED

#include <string>
#include <vector>
#include <map>

using namespace std;

//=============================================================================
//            classes    m e s s a g e   and    m s g _ d a t a
//=============================================================================
// a message is defined by a key (free string), a type (error, warning, ...)
// and a format (string) :
//             text %s text %i text %f ....
// %s for a string data, %i for an int data, %f for a float data
// it supports int, double and string data with no limits
//
// the class msg_data encapsulates the data associated to a message
// a simple way to create a msg_data is to used constructor and then flux operator << :
//          msg_data MD("toto"); MD<<3<<7.5<<"TITI"<<8;
// the main function to handle a message is :   message_event(message_key,message_data)
// for a unique data message : message_event(message_key,int)
//                             message_event(message_key,double)
//                             message_event(message_key,string) or message_event(message_key,char *)
// when an error type message is handle, the program stops.



enum msg_type{error=0,warning,information,question,other_m};
enum language{francais,english,other_l};
typedef unsigned int uint;

class msg
{public :
 static language lang;  //language of the message (default french)
 static string type[4];
 string key;            //key index of the message
 msg_type mt;           //type of the message (error, warning, ...)
 string mes;            //text and related data types
 msg(): mt(other_m) {}
 msg(const string & k, msg_type t,const string &m) : key(k),mt(t),mes(m) {}
};

class list_of_msg
{public :
 map<string,msg> liste;   //list of the messages [user key,message]
 void add(const string& key,msg_type t, const string & m); //add a new message
 msg & operator [](const string &k)  {return liste[k];}
 map<string,msg>::const_iterator find(const string & k) {return liste.find(k);}
 map<string,msg>::const_iterator end() {return liste.end();}
};

//to store msg data
class msg_data
{public :
 vector<int> i_data;    //to store int data
 vector<double> f_data; //to store float data
 vector<string> s_data; //to store string data
 msg_data(){}
 msg_data(int i){i_data.push_back(i);}
 msg_data(double f){f_data.push_back(f);}
 msg_data(const string & s){s_data.push_back(s);}
 msg_data(const char * s){s_data.push_back(string(s));}

 friend msg_data & operator <<(msg_data & d,int n){d.i_data.push_back(n);return d;}
 friend msg_data & operator <<(msg_data & d,double f){d.f_data.push_back(f);return d;}
 friend msg_data & operator <<(msg_data & d,const string & s){d.s_data.push_back(s);return d;}
 friend msg_data & operator <<(msg_data & d,const char * s){d.s_data.push_back(string(s));return d;}
};


//=============================================================================
//                       u t i l i t i e s
//=============================================================================

void push_context(string s);
void pop_context();
void init_messages(language);
void init_french_messages();
void init_english_messages();
void message_event(const string &,const msg_data & d=msg_data());
void message_event(const string &,const string &);
void message_event(const string &,const char *);
void message_event(const string &,double );
void message_event(const string &,int );

#endif // MESSAGE_H_INCLUDED
