#include "utils.h"
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include "message.h"

//size of file
int FileSize(const char* FileName)
{ ifstream f;
  f.open(FileName, std::ios_base::binary | std::ios_base::in);
  if (!f.good() || f.eof() || !f.is_open()) { return 0; }
  f.seekg(0, ios_base::beg);
  ifstream::pos_type begin_pos = f.tellg();
  f.seekg(0,ios_base::end);
  ifstream::pos_type end_pos = f.tellg();
  f.close();
  return static_cast<int>(end_pos - begin_pos);
}

//does a file exist
bool FileExists(const string &Filename)
{struct stat stFileInfo;
 int intStat;
 intStat = stat(Filename.c_str(),&stFileInfo);
 return (intStat == 0);
}

//replace extension characters (.xxxx by .yyy)
string replace_ext(const string & name,const string& ext)
{string newname=name;
 size_t p=name.find_last_of('.',name.size());
 if(p==string::npos) return newname+"."+ext;
 return newname.replace(p+1,newname.size(),ext);
}

//remove all char c in string s
void remove_char(string &s, char c)
{size_t p=s.find(c);
 while (p!=string::npos) {s.erase(p);p=s.find(c);}
}

//replace a char c by an other char nc in a string (all char c are replaced)
string replace(const string & s,char c,char nc)
{string r=s;
 size_t p=r.find(c);
 while (p!=string::npos) {r[p]=nc;p=r.find(c);}
return r;
}

//remove meaningless words Ws from a string S
//"\nW " -> "\n " ; " W " -> " " ; " W\n" -> " \n" ;
void remove_meaningless_words(string & line,const list<string> & m_words)
{list<string>::const_iterator it;
 for(it=m_words.begin();it!=m_words.end();++it)
    {string W=" "+*it+" "; replace(line,W," ");
     W="\n"+*it+" ";replace(line,W,"\n");
     W=" "+*it+"\n";replace(line,W,"\n");
     //special french case (apostrophe)
      if(*it=="D''" || *it=="L''")
        {W=" "+*it;replace(line,W," ");
         W="\n"+*it;replace(line,W,"\n");
        }
    }
}

//in string s, replace all the string w1 by the string w2
void replace(string &s,const string &w1, const string &w2)
{size_t p=s.find(w1);
 uint l=w1.size();
 while (p!=string::npos)
    {s.replace(p,l,w2);
     p=s.find(w1);
    }
}

//remove leading blank
string erase_leading_blank(const string &s)
{string r=s;
 size_t p=r.find_first_not_of (' ');
 if(p>0) r.erase(0,p);
 return r;
}

//remove leading blank from string s
void remove_leading_blank(string &s)
{size_t p=s.find_first_not_of (' ');
 if(p>0) s.erase(0,p);
}

//remove trailing blank
string erase_trailing_blank(const string &s)
{string r=s;
 size_t p=r.find_last_not_of (' ');
 if(p<r.size()-1) r.erase(p+1);
 return r;
}

//remove multiple spaces (keep only one)
void remove_multispaces(string &s)
{string ns="";
 bool inspace=false;
 for(uint p=0;p<s.size();p++)
    {if(s[p]!=' ') {ns+=s[p];inspace=false;}
     else if(!inspace) {ns+=' ';inspace=true;}
    }

 s=ns;
}

//convert string to int (not robust)
int str2int(const string & s)
{stringstream ss(s);
 int i; ss>>i;
 return i;
}

//convert string to int (not robust)
double str2float(const string & s)
{stringstream ss(s);
 double r; ss>>r;
 return r;
}

//convert int to string
string int2str(int i)
{string s;
 return s+i;
}

//convert double to string
string float2str(double f)
{string s;
 return s+f;
}

// string + int
string operator+(const string & s,int i)
{stringstream o;o << i;
 return s+o.str();
}

// string + uint
string operator+(const string & s,uint i)
{return s+int(i);}

// string + double
string operator+(const string & s,double f)
{stringstream o;o << f;
 return s+o.str();
}

//extract info from a f77 format of integer or real (not complex)
//    (n)Tw(.d)
// n: number of numbers
// T: type of number (I integer, F : real fixed, E : scientific)
// w: length of the field
// d: number of digits
// function returns n,F,w,d
void read_f77_format(const string& f,uint &n,char &T,uint &w,uint &d)
{uint m=f.size();
 //detect format type
 string s;uint l=0;char c;T=' ';
 while(l<m && T==' ')
   {c=f[l];l++;
    if(c=='I'|| c=='F' || c=='E') T=c;
    if(T==' ') s+=c;
   }
 if(T==' ') message_event("err_f77_format",msg_data(f));
 if(s!="") n=str2int(s); else n=0;

 //detect field lenght
 if(T=='I' && l<m) {w=str2int(f.substr(l,string::npos));return;}
 if(l>m+1) message_event("err_f77_format",msg_data(f));
 s="";c=f[l];
 while(l<m && c!='.') {s=s+c;l++;c=f[l];}
 if(s!="") w=str2int(s);
 else message_event("err_f77_format",msg_data(f));
 l++;
 if(l<m) d=str2int(f.substr(l,string::npos));
 else message_event("err_f77_format",msg_data(f));
 return;
}

string ignore_quotation_mark(const string& s)
{string so=replace(s,'"',' ');
 so=erase_leading_blank(so);
 so=erase_trailing_blank(so);
 return so;
}

string force_quotation_mark(const string& s)
{string so=s;
 so=erase_leading_blank(so);
 so=erase_trailing_blank(so);
 if(so[0]!='"') so='"'+so;
 if(so[so.size()-1]!='"') so=so+'"';
 return so;
}

vector<double> vect(double x,double y)
  {vector<double> V(2);V[0]=x;V[1]=y;return V;}
vector<double> vect(double x,double y,double z)
  {vector<double> V(3);V[0]=x;V[1]=y;V[2]=z;return V;}


