//=============================================================================
//                  v t k   m e s h   u t i l i t i e s
//=============================================================================
/*
DATASET UNSTRUCTURED_GRID
POINTS nb_point datatype (bit, unsigned_char, char, unsigned_short, short, unsigned_int, int,
x1 y1 z1                            unsigned_long, long, float, or double)
x2 y2 z2
...
CELLS nb_cells size
nbnode n1 n2 ....
...
CELL_TYPES nb_cells
t1
t2
...


WARNING : vtk format is designed only for first order element !!!!!
=============================================================================*/


#include "mesh.h"

extern map<uint,element_type> element_type_list;
extern map<string,mesh_type> mesh_type_list;
extern ofstream debug;


//vtk mesh loading
void mesh::loadfrom_vtk(const char * filename)
{push_context("mesh::loadfrom_vtk");
 message_event("err_filein2",msg_data("vtk"));
 pop_context();
 }

//save mesh to vtk format file
//choice = -1 : asks to the user
//choice = 0  : only the main domain (not the boundary domains) in one file
//choice = 1  : all the domains in different files
void mesh::saveto_vtk(const char * filename,int choice)
{push_context("mesh::saveto_vtk");
 string eol="\n";
//first order element compatibility
 for(cit_le it=elements_list.begin();it!=elements_list.end();++it)
     if(element_type_list[it->second.type].order>1) message_event("err_filein3",msg_data("vtk"));

 //multiple domain analysis (vtk does not manage few domains)
 uint rep_dom=0;
 if(domains_list.size()>1)
    {string ld="";
     map<uint,physical_domain>::const_iterator it;
     for(it=domains_list.begin();it!=domains_list.end();++it) ld+=(it->second.name)+" ";
     if(choice==-1)
       {message_event("que_vtk_domain",msg_data(ld));
        cin>>rep_dom;
       }
     else rep_dom=choice;
    }

 //export only main domain, i.e all elements having the max dimension
 if(rep_dom==0)
 {//open file
 string savename=name;
 if(filename!=NULL) savename=string(filename);
 if(!valid_savefile(savename,string("vtk"))) return;
 ofstream f(savename.c_str(),ios::out | ios::binary);
 if(f.fail()) message_event("open_fail",msg_data(savename));

 //print header
 f<<"# vtk DataFile Version 2.0";
 f<<eol<<erase_trailing_blank(title)<<" (generated by convmesh from "<<name<<")";
 f<<eol<<"ASCII";
 f<<eol<<"DATASET UNSTRUCTURED_GRID";

 //print coordinates of nodes
 f<<eol<<"POINTS "<<nodes_list.size()<<" double";
 for(cit_ln it=nodes_list.begin();it!=nodes_list.end();++it)
    {const node &N=it->second;
     f<<eol<<N(1)<<" "<<N(2)<<" "<<N(3);
    }
 //print elements (only elements of dim = mesh.dim)
 uint n=0,l=0;
 for(cit_le it=elements_list.begin();it!=elements_list.end();++it)
    if(element_type_list[it->second.type].dim==dim)
      {n++;l+=it->second.nodes.size()+1;}
 f<<eol<<eol<<"CELLS "<<n<<" "<<l;
 for(cit_le it=elements_list.begin();it!=elements_list.end();++it)
   if(element_type_list[it->second.type].dim==dim)
     {uint n=it->second.nodes.size();;
      f<<eol<<n;
      for(uint i=0;i<n;i++) f<<" "<<it->second.nodes[i]-1;  //vtk indexes node from 0
     }
 //print element types
 f<<eol<<eol<<"CELL_TYPES "<<n;
 for(cit_le it=elements_list.begin();it!=elements_list.end();++it)
    if(element_type_list[it->second.type].dim==dim)
        f<<eol<<element_type_list[it->second.type].vtk_index;
 //close file
 f.close();
 pop_context();return;
 }

 //export all domains in different files
 map<uint,physical_domain>::const_iterator it;
 for(it=domains_list.begin();it!=domains_list.end();++it)
    {const physical_domain & pd=it->second;
     string savename=name;
     if(filename!=NULL) savename=string(filename);
     size_t p=savename.find_last_of('.');
     if(p==string::npos) message_event("err_filein5",msg_data(savename));

     savename.insert(p,string("_")+pd.name);
     ofstream f(savename.c_str(),fstream::out);
     if(f.fail()) message_event("open_fail",msg_data(savename));
     //loop on the element, re_indexing node
     uint nbnode=0;
     map<uint,uint> index;uint nbn=0;  //index
     vector<uint> r_index;             //reverse index
     //debug<<"\n\n Index pour "<<pd.name;
     for(uint i=0;i<pd.elements.size();i++)
        {const element & e=elements_list(pd.elements[i]);
         nbnode+=e.nodes.size()+1;
         //debug<<e;
         for(uint k=0;k<e.nodes.size();k++)
            {uint j=e.nodes[k];
             map<uint,uint>::iterator it=index.find(j);
             if(it==index.end())
               {nbn++;
                index[j]=nbn;
                r_index.push_back(j);
                //debug<<"\n new index : "<<j<<" "<<nbn;
               }
            }
        }
     //print header
    f<<"# vtk DataFile Version 2.0";
    f<<eol<<erase_trailing_blank(title)<<" domain "<<pd.name<<" (generated by convmesh from "<<name<<")";
    f<<eol<<"ASCII";
    f<<eol<<"DATASET UNSTRUCTURED_GRID";
    //print coordinates of nodes
    f<<eol<<"POINTS "<<index.size()<<" double";
    map<uint,uint>::iterator it;
    for(uint i=0;i<r_index.size();i++)
       {const node &N=nodes_list(r_index[i]);
        //debug<<"\n noeud "<<i<<" -> "<<r_index[i]<<" "<<N;
        f<<eol<<N(1)<<" "<<N(2)<<" "<<N(3);
       }
    //print elements
     uint n=pd.elements.size();
     f<<eol<<eol<<"CELLS "<<n<<" "<<nbnode;
     for(uint i=0;i<n;i++)
        {const element & e=elements_list(pd.elements[i]);
         uint ne=e.nodes.size();
         f<<eol<<ne;
         for(uint j=0;j<ne;j++) f<<" "<<index[e.nodes[j]]-1;  //vtk indexes node from 0
        }
    //print element types
    f<<eol<<eol<<"CELL_TYPES "<<n;
    for(uint i=0;i<n;i++)
        {const element & e=elements_list(pd.elements[i]);
        f<<eol<<element_type_list[e.type].vtk_index;
        }
    //close file
    f.close();
   }
 //end of export
 pop_context();
}
